# -*- coding: utf-8 -*-
#
# miditk/smf/writer.py
#
"""Event handler, which writes back MIDI events to a Standard MIDI File."""

from __future__ import absolute_import, unicode_literals

from io import BytesIO

from six import text_type

from ..common.constants import (ACTIVE_SENSING, CHANNEL_PRESSURE, CONTROLLER_CHANGE, COPYRIGHT,
                                CUEPOINT, END_OF_EXCLUSIVE, END_OF_TRACK, ESCAPE_SEQUENCE, FPS_25,
                                INSTRUMENT_NAME, KEY_SIGNATURE, LYRIC, MARKER, META_EVENT,
                                MIDI_CH_PREFIX, MIDI_TIME_CODE, MONO_PRESSURE, NOTE_OFF, NOTE_ON,
                                PITCH_BEND, PROGRAM_CHANGE, SEQUENCE_NAME, SEQUENCE_NUMBER,
                                SEQUENCER_SPECIFIC, SMTP_OFFSET, SONG_CONTINUE,
                                SONG_POSITION_POINTER, SONG_SELECT, SONG_START, SONG_STOP,
                                SYSTEM_EXCLUSIVE, TEMPO, TEXT, TIME_SIGNATURE, TIMING_CLOCK,
                                TRACK_HEADER, TUNING_REQUEST)
from .api import NullMidiEventHandler
from .converters import tobytestr, write_bew, write_varlen

__all__ = ('MidiFileWriter',)


class MidiFileWriter(NullMidiEventHandler):
    """MidiOutFile writes out MIDI events to a Standard MIDi FIle.

    It subclasses :class:`.NullMidiEventHandler`. For the documentation of method params, see the
    corresponding methods there and in its base class :class:`.BaseMidiEventHandler`.

    """

    def __init__(self, fp, encoding='UTF-8'):
        self._file = fp
        self._track_buffer = None
        super(MidiFileWriter, self).__init__(encoding='UTF-8')

    def event_slice(self, slc):
        """Write the event to the current track.

        Correctly inserts a varlen event timestamp too.

        """
        self._write_varlen(self.relative_time)
        self._write(slc)
        self.update_ticks()

    def _write(self, data):
        """Write the next text slice to the raw data."""
        if isinstance(data, text_type):
            data = data.encode()

        if self._track_buffer is not None:
            self._track_buffer.write(data)
        else:
            self._file.write(data)

    def _write_bew(self, value, length=1):
        """Write a value to the file as big endian word."""
        self._write(write_bew(value, length))

    def _write_varlen(self, value):
        """Write a variable length word to the file."""
        self._write(write_varlen(value))

    #####################
    ## File (non-midi) events

    def header(self, format=0, num_tracks=1, tick_division=96, metrical=True,
               fps=FPS_25, frame_resolution=40):
        """Add MIDI file header chunk.

        For params see :meth:`.BaseMidiEventHandler.header`.

        """
        self._write('MThd')
        self._write_bew(6, 4)  # header size
        self._write_bew(format, 2)
        self._write_bew(num_tracks, 2)

        if metrical:
            # timing devision is in pulses per quarter note (ppqn)
            self._write_bew(tick_division & 32767, 2)
        else:
            # timing devision is in frames per second (fps), frame resolution
            self._write_bew(fps, 1)
            self._write_bew(frame_resolution, 1)

    def start_of_track(self, track=None):
        """Handle start of track meta event.

        :param int track: number of track (None = current track + 1)

        """
        if track is None:
            if self.current_track is None:
                track = 0
            else:
                track = self.current_track + 1

        self.current_track = track
        self._track_buffer = BytesIO()

    #####################
    ## Midi events

    def channel_pressure(self, channel, pressure):
        """Handle channel (mono) pressure event."""
        self.event_slice(tobytestr([CHANNEL_PRESSURE + channel, pressure]))

    def controller_change(self, channel, controller, value):
        """Handle controller change event."""
        self.event_slice(tobytestr([CONTROLLER_CHANGE + channel, controller, value]))

    def note_off(self, channel=0, note=0x40, velocity=0x40):
        """Handle note off event."""
        self.event_slice(tobytestr([NOTE_OFF + channel, note, velocity]))

    def note_on(self, channel=0, note=0x40, velocity=0x40):
        """Handle note on event."""
        self.event_slice(tobytestr([NOTE_ON + channel, note, velocity]))

    def pitch_bend(self, channel, value):
        """Handle pitch bend event."""
        msb = (value >> 7) & 0xFF
        lsb = value & 0xFF
        self.event_slice(tobytestr([PITCH_BEND + channel, msb, lsb]))

    def poly_pressure(self, channel=0, note=0x40, pressure=0x40):
        """Handle poly pressure (aftertouch) event."""
        self.event_slice(tobytestr([MONO_PRESSURE + channel, note, pressure]))

    def program_change(self, channel, program):
        """Handle program change event."""
        self.event_slice(tobytestr([PROGRAM_CHANGE + channel, program]))

    #####################
    ## System Exclusive

    def system_exclusive(self, data):
        """Handle system exclusive (sysex) message."""
        sysex_len = write_varlen(len(data) + 1)
        self.event_slice(tobytestr(SYSTEM_EXCLUSIVE) + sysex_len + data +
                         tobytestr(END_OF_EXCLUSIVE))

    #####################
    ## Meta events

    def meta_slice(self, meta_type, data):
        """Write a meta event."""
        if isinstance(data, text_type):
            data = data.encode(self.encoding, errors='surrogateescape')

        self.event_slice(tobytestr([META_EVENT, meta_type]) +
                         write_varlen(len(data)) + data)

    def end_of_track(self, track=None):
        """Handle end of track meta event.

        Writes the track to the buffer.

        """
        track_data = self._track_buffer.getvalue()
        self._track_buffer = None
        # we need to know size of track data.
        eot_slice = write_varlen(self.relative_time) + tobytestr(
            [META_EVENT, END_OF_TRACK, 0])
        self._write(TRACK_HEADER)
        self._write_bew(len(track_data) + len(eot_slice), 4)
        # then write
        self._write(track_data)
        self._write(eot_slice)

    def copyright(self, text):
        """Handle copyright notice meta event."""
        self.meta_slice(COPYRIGHT, text)

    def cuepoint(self, text):
        """Handle cuepoint meta event."""
        self.meta_slice(CUEPOINT, text)

    def instrument_name(self, text):
        """Handle instrument name meta event."""
        self.meta_slice(INSTRUMENT_NAME, text)

    def key_signature(self, sf, mi):
        """Handle key signature meta event."""
        self.meta_slice(KEY_SIGNATURE, tobytestr([sf, mi]))

    def lyric(self, text):
        """Handle lyrics meta event."""
        self.meta_slice(LYRIC, text)

    def marker(self, text):
        """Handle marker meta event."""
        self.meta_slice(MARKER, text)

    def midi_ch_prefix(self, channel):
        """Handle MIDI channel prefix meta event."""
        self.meta_slice(MIDI_CH_PREFIX, tobytestr([channel]))

    def midi_port(self, value):
        """Handle MIDI port meta event."""
        self.meta_slice(MIDI_CH_PREFIX, tobytestr([value]))

    def sequence_name(self, text):
        """Handle sequence/track name meta event."""
        self.meta_slice(SEQUENCE_NAME, text)

    def sequence_number(self, value):
        """Handle sequence number meta event."""
        self.meta_slice(SEQUENCE_NUMBER, write_bew(value, 2))

    def sequencer_specific(self, data):
        """Handle sequencer specific meta event."""
        self.meta_slice(SEQUENCER_SPECIFIC, data)

    def smtp_offset(self, hour, minute, second, frame, frame_part):
        """Handle SMTP offset meta event."""
        self.meta_slice(SMTP_OFFSET,
                        tobytestr([hour, minute, second, frame, frame_part]))

    def tempo(self, value):
        """Handle tempo meta event."""
        hb, mb, lb = (value >> 16 & 0xff), (value >> 8 & 0xff), (value & 0xff)
        self.meta_slice(TEMPO, tobytestr([hb, mb, lb]))

    def text(self, text):
        """Handle text meta event."""
        self.meta_slice(TEXT, text)

    def time_signature(self, nn, dd, cc, bb):
        """Handle time signature meta event."""
        self.meta_slice(TIME_SIGNATURE, tobytestr([nn, dd, cc, bb]))

    def unknown_meta_event(self, meta_type, data):
        """Handle any undefined meta events."""
        self.meta_slice(meta_type, tobytestr(data))

    #####################
    ## Common events
    ##
    ## These should only occur in real-time MIDI data, not in Standard MIDI Files.

    def midi_time_code(self, msg_type, values):
        """Handle MIDI time code (MTC) quarter frame system common event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        :param int msg_type: 0-7
        :param int values: 0-15

        """
        value = (msg_type << 4) + values
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 2, MIDI_TIME_CODE, value]))

    def song_position_pointer(self, value):
        """Handle song position pointer (SPP) system common event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        :param int value: 0-16383

        """
        lsb = (value & 0x7F)
        msb = (value >> 7) & 0x7F
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 3, SONG_POSITION_POINTER, lsb, msb]))

    def song_select(self, song_number):
        """Handle song select system common event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        :param int song_number: 0-127

        """
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 2, SONG_SELECT, song_number]))

    def tuning_request(self):
        """Handle tuning request system common event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        Takes no arguments.

        """
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 1, TUNING_REQUEST]))

    #####################
    ## Real-time events
    ##
    ## These should only occur in real-time MIDI data, not in Standard MIDI Files.

    def timing_clock(self):
        """Handle MIDI timing clock system real-time event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        Takes no arguments.

        """
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 1, TIMING_CLOCK]))

    def song_start(self):
        """Handle song start system real-time event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        Takes no arguments.

        """
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 1, SONG_START]))

    def song_stop(self):
        """Handle song stop system real-time event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        Takes no arguments.

        """
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 1, SONG_STOP]))

    def song_continue(self):
        """Handle song continue system real-time event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        Takes no arguments.

        """
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 1, SONG_CONTINUE]))

    def active_sensing(self):
        """Handle active sensing system real-time event.

        This should only occur in real-time MIDI data, in Standard MIDI Files it must be embedded
        in an escape sequence meta event.

        Takes no arguments.

        """
        self.event_slice(tobytestr([ESCAPE_SEQUENCE, 1, ACTIVE_SENSING]))

    def system_reset(self):
        """Handle system reset system real-time event.

        This can only occur in real-time MIDI data, in Standard MIDI Files 0xFF marks a meta event,
        so it is ignored.

        Takes no arguments.

        """
        pass



# -*- coding:utf-8 -*-
#
# miditk/smf/release.py - release information for the python-midi package
#

name = 'miditk-smf'
version = '0.2.0'
keywords = 'MIDI, parsing, multimedia, I/O'
author = 'Christopher Arndt'
author_email = 'chris@chrisarndt.de'
url = 'https://github.com/SpotlightKid/miditk-smf'
license = 'MIT License'
description = "A toolkit for working with Standard MIDI files"
platforms = 'POSIX, Windows, MacOS X'
classifiers = """
Development Status :: 4 - Beta
Intended Audience :: Developers
Intended Audience :: End Users/Desktop
License :: OSI Approved :: MIT License
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: MacOS :: MacOS X
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: PyPy
Topic :: Software Development :: Libraries :: Python Modules
Topic :: Multimedia :: Sound/Audio :: MIDI
Topic :: Utilities
"""
classifiers = [c.strip() for c in classifiers.splitlines()
               if c.strip() and not c.startswith('#')]

try:
    del c  # noqa
except NameError:
    pass  # Python 3



# -*- coding:utf-8 -*-

from __future__ import print_function, unicode_literals

import sys
from io import StringIO
from os.path import dirname, join

import pytest  # noqa

from miditk.smf.api import PrintingMidiEventHandler
from miditk.smf.reader import MidiFileReader


def test_minimal_type0():
    """Test that we get the right output for a minimal type 0 midi test file."""
    test_file = join(dirname(__file__), 'testdata', 'minimal-cubase-type0.mid')

    # Do parsing, and generate events with MidiToText, so we can see what a
    # minimal midi file contains:

    midi_in = MidiFileReader(test_file, PrintingMidiEventHandler())
    old_stdout = sys.stdout
    capture_stdout = StringIO()
    sys.stdout = capture_stdout
    midi_in.read()
    result = capture_stdout.getvalue()
    sys.stdout = old_stdout
    assert result.strip() == """\
format: 0, no. of tracks: 1, tick division: 480 ppqn
------------------------------------------------------------

Start of track #0
Sequence name - 'Type 0'
Tempo - val:500000 (120.00 bpm)
Time signature - 4/4 24 8
Note on - ch:00, note:48h, vel:64h time:0
Note off - ch:00, note:48h, vel:40h time:480
End of track #0

End of file"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example script which uses the PrintingMidiEventHandler.

When an event is passed to this handler, it prints information about the event to the console.

The events are passed to the handler by the MidiFileReader instance.

This script prints all the events from the infile to the console, which is useful for
debugging.

"""

import sys
from os.path import abspath, dirname, join

from miditk.smf import MidiFileReader
from miditk.smf.api import PrintingMidiEventHandler


try:
    infile = sys.argv[1]
except IndexError:
    infile = join(dirname(dirname(abspath(__file__))), 'tests', 'testdata',
                  'minimal-cubase-type0.mid')

# create the reader and the event handler
midiin = MidiFileReader(infile, PrintingMidiEventHandler())
# do parsing
midiin.read()



# -*- coding: UTF-8 -*-

import pytest

from miditk.smf.converters import read_bew, read_varlen, tobytestr, write_bew, write_varlen


def test_read_bew():
    """Test that 'read_bew' returns correct values."""
    assert read_bew(b'a\xe1\xe2\xe3') == 1642193635
    assert read_bew(b'\x04\x00') == 1024
    assert read_bew(b'\x7f') == 127


def test_read_varlen():
    """Test that 'read_varlen' returns correct values."""
    assert read_varlen(b'\xc0\x00') == 8192
    assert read_varlen(b'\xe1\xe3\xe2a') == 205058401


def test_write_bew():
    """Test that 'read_bew' returns correct values."""
    assert write_bew(1642193635, 4) == b'a\xe1\xe2\xe3'
    assert write_bew(1024, 2) == b'\x04\x00'
    assert write_bew(127, 1) == b'\x7f'


@pytest.mark.parametrize("test_input,expected", [
    (192, b'\x81\x40'),
    (8192, b'\xc0\x00'),
    (0x00000000, b'\0'),
    (0x00000040, b'\x40'),
    (0x0000007F, b'\x7F'),
    (0x00000080, b'\x81\x00'),
    (0x00002000, b'\xC0\x00'),
    (0x00003FFF, b'\xFF\x7F'),
    (0x00004000, b'\x81\x80\x00'),
    (0x00100000, b'\xC0\x80\x00'),
    (0x001FFFFF, b'\xFF\xFF\x7F'),
    (0x00200000, b'\x81\x80\x80\x00'),
    (0x08000000, b'\xC0\x80\x80\x00'),
    (0x0FFFFFFF, b'\xFF\xFF\xFF\x7F'),
])
def test_write_varlen(test_input, expected):
    """Test that 'write_varlen' returns correct values."""
    assert write_varlen(test_input) == expected


def test_tobytestr():
    assert tobytestr((48, 49, 50)) == b'\x30\x31\x32'
    assert tobytestr('\x30\x31\x32') == b'\x30\x31\x32'
    assert tobytestr(u'\x30\x31\x32') == b'\x30\x31\x32'
    assert tobytestr(b'\x30\x31\x32') == b'\x30\x31\x32'


def tointseq():
    assert tointseq('\x30\x31\x32') == (48, 49, 50)
    assert tointseq(u'\x30\x31\x32') == (48, 49, 50)
    assert tointseq(b'\x30\x31\x32') == (48, 49, 50)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Write MIDI file with the following events:

format: 0, # of tracks: 1, division: 480
----------------------------------

Start of track #0
Sequence name - 'Type 0'
Tempo: val:500000 (125.0 bpm)
Time signature: 4/4 24 8
Note on - ch:00, note:48h, vel:64h time:0
Note off - ch:00, note:48h, vel:40h time:480
End of track
End of file

"""

import os
from os.path import dirname, isdir, join

import pytest  # noqa

from miditk.smf.writer import MidiFileWriter


def test_write_type0():
    cmpfn = join(dirname(__file__), 'testdata', 'midiout.mid')
    outdir = join(dirname(__file__), 'testoutput')
    outfn = join(outdir, 'midiout.mid')

    if not isdir(outdir):
        os.makedirs(outdir)

    with open(outfn, 'wb') as smf:
        midi = MidiFileWriter(smf)

        midi.header(0, 1, 480)

        midi.start_of_track()
        midi.sequence_name('Type 0')
        midi.tempo(750000)
        midi.time_signature(4, 2, 24, 8)
        ch = 0

        for i in range(127):
            midi.note_on(ch, i, 0x64)
            midi.update_ticks(96)
            midi.note_off(ch, i, 0x40)
            midi.update_ticks(0)

        midi.update_ticks(0)
        midi.end_of_track()
        midi.eof()

    with open(outfn, 'rb') as out, open(cmpfn, 'rb') as cmp:
        assert cmp.read() == out.read()



# -*- coding: utf-8 -*-
#
# miditk/smf/api.py
#
"""MIDI file parsing event handler classes"""

from __future__ import absolute_import, print_function, unicode_literals

# package-specific imports
from ..common.constants import (CHANNEL_PRESSURE, CONTROLLER_CHANGE, COPYRIGHT, CUEPOINT,
                                DEVICE_NAME, END_OF_TRACK, INSTRUMENT_NAME, KEY_SIGNATURE, LYRIC,
                                MARKER, MIDI_CH_PREFIX, MIDI_PORT, NOTE_OFF, NOTE_ON, PITCH_BEND,
                                POLYPHONIC_PRESSURE, PROGRAM_CHANGE, PROGRAM_NAME, SEQUENCE_NAME,
                                SEQUENCE_NUMBER, SEQUENCER_SPECIFIC, SMTP_OFFSET, TEMPO, TEXT,
                                TIME_SIGNATURE)
from .converters import read_bew, tointseq


__all__ = (
    'BaseMidiEventHandler',
    'NullMidiEventHandler',
    'PrintingMidiEventHandler'
)


class BaseMidiEventHandler(object):
    """Dispatch events to specific event handler methods.

    Analyses MIDI events (i.e. channel message and meta events) and calls
    appropriate methods according to the type of the event on the handler
    instance.

    Also keeps track of event times and maintains a running absolute time
    and a relative time between events. Time values are in absolute
    tick values from the opening of a stream.

    This class must be subclassed by a concrete event handler class, which
    implements all the methods to handle specific types of events.

    Sysex and invalid events are silently ignored by default, overwrite the
    ``sysex_event`` and ``invalid_event`` methods in a subclass to handle them.

    """

    def __init__(self, encoding='UTF-8'):
        """Initialize event dispatcher with given output event handler."""
        self.encoding = encoding
        self.absolute_time = 0
        self.relative_time = 0
        self.current_track = None

        # public flags

        # A note_on with a velocity of 0x00 is actually the same as a
        # note_off with a velocity of 0x40. When
        # "convert_zero_velocity" is set, the zero velocity note_on's
        # automatically gets converted into note_off's. This is a less
        # suprising behaviour for those that are not into the intimate
        # details of the midi spec.
        self.convert_zero_velocity = True
        self.dispatch_controllers = False

    ## Handle parsing events, i.e. non-midi, file structure and timing events

    #############################
    ## MIDI file header handler

    def header(self, format, num_tracks, tick_division=96, metrical=True,
               fps=0xE7, frame_resolution=0x28):
        """Handle file header event.

        :param int format: type of midi file in [0, 1, 2]
        :param int num_tracks: number of tracks. Must be 1 track for type 0 file
        :param int tick_division: timing division, i.e. 96 ppqn.
        :param bool metrical: whether to use metrical timing (tick_division) or timecode
            (fps / frame_resolution)
        :param int fps: frames per second as negative 2-complement
        :param int frame_resolution: number of subframes per frame

        """
        pass

    #############################
    ## Parsing (non-midi) events

    def eof(self):
        """Handle end of file event.

        Does nothing by default and can be safely overwritten in a subclass.

        """
        self.current_track = None

    def start_of_track(self, track):
        """Handle start of track event.

        Saves the track number in the instance attribute ``current_track``
        for use by the other event handler methods.

        """
        self.current_track = track

    def reset_ticks(self):
        """Reset time to zero.

        If you overwite this in a sub class, be sure to call this method from
        the base class.

        """
        self.relative_time = 0
        self.absolute_time = 0

    def update_ticks(self, ticks=0, relative=True):
        """Update the global time.

        If ``relative`` is True, ``ticks`` is relative, else it's absolute.

        If you overwite this in a sub class, be sure to call this method from
        the base class.

        """
        if relative:
            self.relative_time = ticks
            self.absolute_time += ticks
        else:
            self.relative_time = ticks - self.absolute_time
            self.absolute_time = ticks

    #############################
    ## Track events

    ## Ignored events

    def invalid_event(self, event):
        """Handle invalid system common or system real time messages.

        Does nothing by default and can be safely overwritten in a subclass.

        :param event: :class:`.MidiEvent` with 0xF1 <= event.type <= 0xFE

        """
        pass

    def escape_sequence(self, event):
        """Handle escape sequence.

        Does nothing by default and can be safely overwritten in a subclass.

        :param event: :class:`.MidiEvent` of type ``ESCAPE_SEQUENCE``

        """
        pass

    # Dispatchers for sysex, channel and meta events

    def sysex_event(self, event):
        """Handle sysex events.

        Calls ``system_exclusive`` method on the handler and passes the message data as a byte
        string.

        :param event: :class:`.MidiEvent` of type ``SYSTEM_EXCLUSIVE``

        """
        self.sysex_message(event.data)

    def channel_message_event(self, event):
        """Dispatch channel message events.

        Examines the message type, parses the event data according to type and
        calls an appropriate event handler method with the event data.

        Type and number of arguments of the handler methods vary depending on
        the event type.

        :param event: :class:`.MidiEvent` with 0x80 <= event.type <= 0xE0

        """
        data = tointseq(event.data)

        if event.type == NOTE_ON:
            note, velocity = data
            # note_on with velocity 0x00 are same as note
            # off with velocity 0x40 according to spec!
            if velocity == 0 and self.convert_zero_velocity:
                self.note_off(event.channel, note, 0x40)
            else:
                self.note_on(event.channel, note, velocity)

        elif event.type == NOTE_OFF:
            note, pressure = data
            self.note_off(event.channel, note, pressure)

        elif event.type == POLYPHONIC_PRESSURE:
            note, pressure = data
            self.poly_pressure(event.channel, note, pressure)

        elif event.type == CONTROLLER_CHANGE:
            controller, value = data
            self.controller_change(event.channel, controller, value)

        elif event.type == PROGRAM_CHANGE:
            program = data[0]
            self.program_change(event.channel, program)

        elif event.type == CHANNEL_PRESSURE:
            pressure = data[0]
            self.channel_pressure(event.channel, pressure)

        elif event.type == PITCH_BEND:
            hibyte, lobyte = data
            value = (hibyte << 7) + lobyte
            self.pitch_bend(event.channel, value)

        else:
            self.unknown_channel_message(event.type, data)

    def meta_event(self, event):
        """Dispatch meta events.

        Examines the message type, parses the event data according to type and
        calls an appropriate event handler method with the event data.

        Type and number of arguments of the handler methods vary depending on
        the event type.

        :param event: :class:`.MidiEvent` of type <= ``META_EVENT``

        """
        data = event.data

        # Events sorted by most common first for optimization

        # END_OF_TRACK = 0x2F (2F 00)
        if event.meta_type == END_OF_TRACK:
            self.end_of_track(self.current_track)
            self.current_track = None

        # TEMPO = 0x51 (51 03 tt tt tt (tempo in us/quarternote))
        elif event.meta_type == TEMPO:
            b1, b2, b3 = tointseq(data)
            # uses 3 bytes to represent time between quarter
            # notes in microseconds
            self.tempo((b1 << 16) + (b2 << 8) + b3)

        # TIME_SIGNATURE = 0x58 (58 04 nn dd cc bb)
        elif event.meta_type == TIME_SIGNATURE:
            nn, dd, cc, bb = tointseq(data)
            self.time_signature(nn, dd, cc, bb)

        # KEY_SIGNATURE = 0x59 (59 02 sf mi)
        elif event.meta_type == KEY_SIGNATURE:
            sf, mi = tointseq(data)
            self.key_signature(sf, mi)

        # SEQUENCE_NAME = 0x03 (03 len text...)
        elif event.meta_type == SEQUENCE_NAME:
            self.sequence_name(data)

        # PROGRAM_NAME = 0x08 (05 len text...)
        elif event.meta_type == PROGRAM_NAME:
            self.program_name(data)

        # INSTRUMENT_NAME = 0x04 (04 len text...)
        elif event.meta_type == INSTRUMENT_NAME:
            self.instrument_name(data)

        # TEXT = 0x01 (01 len text...)
        elif event.meta_type == TEXT:
            self.text(data)

        # COPYRIGHT = 0x02 (02 len text...)
        elif event.meta_type == COPYRIGHT:
            self.copyright(data)

        # SEQUENCE_NUMBER = 0x00 (00 02 ss ss (seq-number))
        elif event.meta_type == SEQUENCE_NUMBER:
            number = read_bew(data)
            self.sequence_number(number)

        # LYRIC = 0x05 (05 len text...)
        elif event.meta_type == LYRIC:
            self.lyric(data)

        # MARKER = 0x06 (06 len text...)
        elif event.meta_type == MARKER:
            self.marker(data)

        # CUEPOINT = 0x07 (07 len text...)
        elif event.meta_type == CUEPOINT:
            self.cuepoint(data)

        # DEVICE_NAME = 0x09 (09 len text...)
        elif event.meta_type == DEVICE_NAME:
            self.device_name(data)

        # MIDI_CH_PREFIX = 0x20 (20 01 channel)
        elif event.meta_type == MIDI_CH_PREFIX:
            channel = read_bew(data)
            self.midi_ch_prefix(channel)

        # MIDI_PORT  = 0x21 (21 01 port (legacy stuff))
        elif event.meta_type == MIDI_PORT:
            port = read_bew(data)
            self.midi_port(port)

        # SMTP_OFFSET = 0x54 (54 05 hh mm ss ff xx)
        elif event.meta_type == SMTP_OFFSET:
            hour, minute, second, frame, framePart = tointseq(data)
            self.smtp_offset(hour, minute, second, frame, framePart)

        # SEQUENCER_SPECIFIC = 0x7F (Sequencer specific event)
        elif event.meta_type == SEQUENCER_SPECIFIC:
            meta_data = tointseq(data)
            self.sequencer_specific(meta_data)

        # Handles any undefined meta events
        else:  # undefined meta type
            meta_data = tointseq(data)
            self.unknown_meta_event(event.meta_type, meta_data)


class NullMidiEventHandler(BaseMidiEventHandler):
    """NullMidiEventHandler handles all MIDI events by doing nothing.

    It can serve as a base class for your specific MIDI event handler classes
    where you only want to handle some events and ignore all others.

    """

    #####################
    ## Channel mesages

    def channel_pressure(self, channel, pressure):
        """Handle channel (mono) change message.

        :param int channel: 0-15
        :param int pressure: 0-127

        """
        pass

    def controller_change(self, channel, controller, value):
        """Handle controller change message.

        :param int channel: 0-15
        :param int controller: 0-127
        :param int value: 0-127

        """
        pass

    def note_off(self, channel=0, note=0x40, velocity=0x40):
        """Handle Note off message.

        :param int channel: 0-15
        :param int note: 0-127
        :param int velocity: 0-127

        """
        pass

    def note_on(self, channel=0, note=0x40, velocity=0x40):
        """Handle Note on message.

        :param int channel: 0-15
        :param int note: 0-127
        :param int velocity: 0-127

        """
        pass

    def poly_pressure(self, channel=0, note=0x40, velocity=0x40):
        """Handle poly pressure message.

        :param int channel: 0-15
        :param int note: 0-127
        :param int velocity: 0-127

        """
        pass

    def pitch_bend(self, channel, value):
        """Handle pitch bend message.

        :param int channel: 0-15
        :param int value: 0-16383

        """
        pass

    def program_change(self, channel, program):
        """Handle program change message.

        :param int channel: 0-15
        :param int program: 0-127

        """
        pass

    def unknown_channel_message(self, status, data):
        """Handle unrecognized channel messages."""
        pass

    #####################
    ## System exclusive events

    def system_exclusive(self, data):
        """Handle system exclusive messsage.

        :param bytes data: byte string of values in range(128)

        """
        pass

    #####################
    ## Meta events

    def copyright(self, text):
        """Handle copyright notice event.

        :param str text: copyright text

        """
        pass

    def cuepoint(self, text):
        """Handle cue point event.

        :param str text: cue point text

        """
        pass

    def device_name(self, text):
        """Handle device name event.

        :param str text: device name

        """
        pass

    def end_of_track(self, track=None):
        """Handle end of track event.

        :param int track: number of track (None = current track)

        """
        pass

    def instrument_name(self, text):
        """Handle instrument name event.

        :param str text: instrument name

        """
        pass

    def key_signature(self, sf, mi):
        """Handle key signature event.

        :param int sf: a byte specifying the number of flats (-ve) or sharps (+ve) which
          identifies the key signature (-7 = 7 flats, -1 = 1 flat, 0 = key of C, 1 = 1 sharp, etc).
        :param int mi: a byte specifying a major (0) or minor (1) key.

        """
        pass

    def lyric(self, text):
        """Handle lyric event.

        :param str text: lyric text

        """
        pass

    def marker(self, text):
        """Handle text marker event.

        :param str text: marker text

        """
        pass

    def midi_ch_prefix(self, channel):
        """Handle (deprecated) MIDI channel prefix event.

        :param int channel: midi channel number for subsequent data

        """
        pass

    def midi_port(self, value):
        """Handle (deprecated) MIDI port event.

        :param int value: midi port number (deprecated)

        """
        pass

    def sequence_name(self, text):
        """Handle sequence / track name event.

        In a Type-1 MIDI file, the track name event of track #0 should be interpreted as the
        name of the sequence as a whole.

        param:str:text: sequence / track name

        """
        pass

    def sequence_number(self, value):
        """Handle sequence number event.

        :param int value: sequence number (0-16383)

        """
        pass

    def sequencer_specific(self, data):
        """Handle sequencer specific event.

        :param bytes data: The data as byte values

        """
        pass

    def smtp_offset(self, hour, minute, second, frame, frame_part):
        """Handle SMTP offset event.

        :param int hour: the hour (0-23). The hour should be encoded with the SMPTE format, just as
          it is in MIDI Time Code.
        :param int minute: minutes (0-59)
        :param int second: seconds (0-59)
        :param int frame: a byte specifying the number of frames per second (one of:
            24, 25, 29, 30).
        :param int frame_part: a byte specifying the number of fractional frames, in 100ths of a
            frame (even in SMPTE-based tracks using a different frame subdivision, defined in the
            MThd chunk).

        """
        pass

    def tempo(self, value):
        """Handle tempo event.

        The tempo es given in in µs/quarternote.

        :param int value: 0-2097151

        To calculate the tempo value from bpm: ``int(60,000,000 / BPM)``

        """
        pass

    def text(self, text):
        """Handle text event.

        :param str text: arbitrary text

        """
        pass

    def time_signature(self, nn, dd, cc, bb):
        """Handle time signature event.

        :param int nn: Numerator of the signature as notated on sheet music
        :param int dd: Denominator of the signature as notated on sheet music. The denominator is a
          negative power of 2: 2 = quarter note, 3 = eighth, etc.
        :param int cc: The number of MIDI clocks in a metronome click
        :param int bb: The number of notated 32nd notes in a MIDI quarter note (24 MIDI clocks)

        """
        pass

    def unknown_meta_event(self, meta_type, data):
        """Handle any undefined meta events."""
        pass


class PrintingMidiEventHandler(NullMidiEventHandler):
    """This class reports basic information about MIDI events to standard
    output as they are encountered.

    It is mostly useful for debugging.

    """

    #########################
    ## File events

    def header(self, format=0, num_tracks=1, tick_division=96, metrical=True,
               fps=0xE7, frame_resolution=0x28):
        if metrical:
            print('format: %s, no. of tracks: %i, tick division: %i ppqn' %
                  (format, num_tracks, tick_division))
        else:
            print('format: %s, no. of tracks: %i, fps: %i, resolution: %i' %
                  (format, num_tracks, fps, frame_resolution))
        print('-' * 60)
        print('')

    def eof(self):
        print('End of file')

    def start_of_track(self, track=0):
        super(PrintingMidiEventHandler, self).start_of_track(track)
        print('Start of track #%s' % track)

    #############################
    ## Channel events

    def channel_pressure(self, channel, pressure):
        print('Channel pressure - ch:%02i, pressure=%02Xh' %
              (channel, pressure))

    def controller_change(self, channel, controller, value):
        print('Controller - ch: %02i, cont:%02Xh, value:%02Xh' %
              (channel, controller, value))

    def note_off(self, channel=0, note=0x40, velocity=0x40):
        print('Note off - ch:%02i, note:%02Xh, vel:%02Xh time:%s' %
              (channel, note, velocity, self.relative_time))

    def note_on(self, channel=0, note=0x40, velocity=0x40):
        print('Note on - ch:%02i, note:%02Xh, vel:%02Xh time:%s' %
              (channel, note, velocity, self.relative_time))

    def pitch_bend(self, channel, value):
        print('Pitch bend - ch:%02s, value:%02Xh' % (channel, value))

    def poly_pressure(self, channel=0, note=0x40, pressure=0x40):
        print('Poly pressure - ch: %02i, note:%02Xh, pressure:%02Xh',
              (channel, note, pressure))

    def program_change(self, channel, program):
        print('Program change - ch:%02i, program:%02Xh' % (channel, program))

    def unknown_channel_message(self, status, data):
        print('Unknown channel message - status:%s, data:%r' % (status, data))

    ###############
    ## sysex event

    def system_exclusive(self, data):
        print('System exclusive message - size:%i' % (len(data) + 1))

    #####################
    ## meta events

    def copyright(self, text):
        print("Copyright - '%s'" % text.decode(self.encoding))

    def cuepoint(self, text):
        print("Cuepoint - '%s'" % text.decode(self.encoding))

    def device_name(self, text):
        print("Device name - '%s'" % text.decode(self.encoding))

    def end_of_track(self, track=None):
        print('End of track #%s' % track)
        print('')

    def instrument_name(self, text):
        print("Instrument name - '%s'" % text.decode(self.encoding))

    def key_signature(self, sf, mi):
        print('Key signature - ' % (sf, mi))

    def marker(self, text):
        print("Marker - '%s'" % text.decode(self.encoding))

    def midi_ch_prefix(self, channel):
        print('MIDI channel prefix - ch:%02i' % channel)

    def midi_port(self, port):
        print('MIDI port - port:%i' % port)

    def lyric(self, text):
        print("Lyric - '%s'" % text.decode(self.encoding))

    def program_name(self, text):
        print("Program name - '%s'" % text.decode(self.encoding))

    def sequence_name(self, text):
        print("Sequence name - '%s'" % text.decode(self.encoding))

    def sequence_number(self, number):
        print('Sequence number - no.:%i' % number)

    def sequencer_specific(self, data):
        print('Sequencer specific - size:%i' % len(data))

    def smtp_offset(self, hour, minute, second, frame, framePart):
        print('SMTP offset - %02i:%02i:%02i.%i/%i' %
              (hour, minute, second, frame, framePart))

    def tempo(self, value):
        print('Tempo - val:%i (%.2f bpm)' % (value, 60000000.00 / value))

    def text(self, text):
        print("Text - '%s'" % text.decode(self.encoding))

    def time_signature(self, nn, dd, cc, bb):
        print('Time signature - %i/%i %i %i' % (nn, 2**dd, cc, bb))

    def unknown_meta_event(self, meta_type, data):
        print('Unknown meta event - type:%s, size:%i' %
              (meta_type, len(data)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example script which creates the smallest possible type 0 midi file.

All the midi events are in the same track chunk, which only contains one note
on and off event.

"""

import sys
from os.path import abspath, dirname, join

from miditk.smf import MidiFileWriter


try:
    outfile = sys.argv[1]
except IndexError:
    outfile = join(dirname(dirname(abspath(__file__))), 'tests', 'testoutput', 'minimal-type0.mid')

with open(outfile, 'wb') as smf:
    midi = MidiFileWriter(smf)

    # write file and track header
    midi.header()
    midi.start_of_track()

    # musical events
    midi.note_on(channel=0, note=0x40)

    midi.update_ticks(192)
    midi.note_off(channel=0, note=0x40)

    # end track and midi file
    midi.end_of_track()
    midi.eof()



# -*- coding: utf-8 -*-
#
# miditk/smf/reader.py
#
"""MidiFileReader combines the MidiFileParser with an event handler."""

from __future__ import absolute_import, unicode_literals

from io import BytesIO

from six import string_types, text_type

# package-specific imports
from .api import NullMidiEventHandler
from .parser import MidiFileParser


__all__ = ('MidiFileReader',)


class MidiFileReader(object):
    """Parse a midi file and trigger the midi events on the outstream object."""

    def __init__(self, infile, handler=None):
        self.infile = infile
        self.handler = handler

    def read(self, strict=True):
        """Start parsing the file."""
        if isinstance(self.infile, string_types):
            self.infile = open(self.infile, 'rb')
            should_close = True
        else:
            should_close = False

        self.parser = MidiFileParser(self.infile, self.handler or NullMidiEventHandler,
                                     strict=strict)
        self.parser.parse_header()
        self.parser.parse_tracks()

        if should_close:
            self.infile.close()

    def set_data(self, data='', encoding='UTF-8'):
        """Set parser data from a plain or byte string."""
        if isinstance(data, text_type):
            data = data.decode(encoding)

        try:
            self.infile.close()
        except (OSError, IOError):
            pass

        self.infile = BytesIO(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example script which transposes all note on/off event by one octave."""

from os.path import abspath, dirname, join
from miditk.smf import MidiFileReader, MidiFileWriter


# MidiFileWriter is a sub-class of NullMidiEventHandler
class Transposer(MidiFileWriter):
    """Transpose note values of all note on/off events by 1 octave."""

    def _transp(self, ch, note):
        # don't transpose note son teh drum channel
        if ch != 9:
            note = min(127, note + 12)

        return note

    def note_on(self, channel, note, velocity):
        note = self._transp(channel, note)
        super(Transposer, self).note_on(channel, note, velocity)

    def note_off(self, channel, note, velocity):
        note = self._transp(channel, note)
        super(Transposer, self).note_off(channel, note, velocity)


# This is a small type 0 midi file, with only two midi events on track #0.
infile = join(dirname(dirname(abspath(__file__))), 'tests', 'testdata', 'minimal.mid')
# The output is written to the tests/testoutput directory
outfile = join(dirname(dirname(abspath(__file__))), 'tests', 'testoutput', 'transposed.mid')

# create parsre and event handler
with open(outfile, 'wb') as smf:
    midiout = Transposer(smf)
    midiin = MidiFileReader(infile, midiout)
    # now do the processing
    midiin.read()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging

from .converters import (read_varlen, read_bew, sizeof_varlen, tobytestr,  # noqa: F401
                         tointseq, write_bew, write_varlen)  # noqa: F401
from .api import BaseMidiEventHandler, NullMidiEventHandler  # noqa: F401
from .parser import MidiFileParser  # noqa: F401
from .reader import MidiFileReader  # noqa: F401
from .release import version as __version__  # noqa: F401
from .sequence import MidiSequence  # noqa: F401
from .writer import MidiFileWriter  # noqa: F401


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


logging.getLogger(__name__).addHandler(NullHandler())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# setup.py - Setup file for the miditk-smf package
#

from __future__ import print_function

import distutils
import subprocess
import sys

from os.path import dirname, join

from setuptools import setup


def read(*args):
    return open(join(dirname(__file__), *args)).read()


class ToxTestCommand(distutils.cmd.Command):
    """Distutils command to run tests via tox with 'python setup.py test'.

    Please note that in this configuration tox uses the dependencies in
    `requirements/dev.txt`, the list of dependencies in `tests_require` in
    `setup.py` is ignored!

    See https://docs.python.org/3/distutils/apiref.html#creating-a-new-distutils-command
    for more documentation on custom distutils commands.

    """
    description = "Run tests via 'tox'."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.announce("Running tests with 'tox'...", level=distutils.log.INFO)
        return subprocess.call(['tox'])


install_requires = [
    'six',
]

tests_require = [
    # intentionally left empty, test requirements are declared in tox.ini!
]

scripts = [
    join("miditk", "smf", "scripts", "mid2syx.py")
]

if sys.platform.startswith('win') and 'py2exe' in sys.argv:
    try:
        import py2exe  # noqa: F401
    except ImportError:
        print("py2exe not found. Cannot build Windows stand-alone executables "
              "without it. Please install py2exe.", file=sys.stderr)
        sys.exit(1)
    else:
        setup_opts = dict(
            options={'py2exe': {'bundle_files': 1}},
            console=scripts,
            zipfile=None
        )
else:
    setup_opts = dict(
        entry_points={
            'console_scripts': [
                'mid2syx = miditk.smf.scripts.mid2syx:main',
            ]
        },
        cmdclass={
            'test': ToxTestCommand,
        },
        zip_safe=False
    )


# read meta-data from miditk/smf/release.py
exec(read('miditk', 'smf', 'release.py'), {}, setup_opts)


setup(
    long_description=read('README.rst'),
    packages=[
        'miditk',
        'miditk.common',
        'miditk.smf',
        'miditk.smf.scripts'
    ],
    include_package_data=True,
    test_suite='tests',
    install_requires=install_requires,
    tests_require=tests_require,
    **setup_opts
)
